/*
 * https://github.com/mattma/angular-cli-generator
 *
 * Copyright (c) 2018 Matt Ma
 * Licensed under the MIT license.
 */

'use strict';

const program = require('commander');
const path = require('path');
const fs = require('fs');
const gulp = require('gulp');
const gutil = require('gulp-util');
const replace = require('gulp-replace');
const rename = require('gulp-rename');
const stringUtils = require('../utils/string');

const VALID_TYPES = ['route'];

// Modify app routing: app-routing.module.ts
function updateRouting(options) {
  const { moduleName, project } = options;
  const srcRootPath = project ? `projects/${project}/src/app` : 'src/app';
  const routingPath = `${srcRootPath}/app-routing.module.ts`;
  const dasherizeName = stringUtils.dasherize(moduleName);
  const classifyName = stringUtils.classify(moduleName);

  gulp
    .src(routingPath)
    .pipe(
      replace(
        /{\s+path\:\s+'not-found-404',/g,
        `{
    path: '${moduleName}',
    loadChildren: '${srcRootPath}/${dasherizeName}/${dasherizeName}.module#${classifyName}Module',
  },
  {
    path: 'not-found-404',`
      )
    )
    .on('end', () => {
      gutil.log(gutil.colors.yellow('  update'), routingPath);
    })
    .pipe(gulp.dest(srcRootPath));
}

// Core function to generate the file from template, insert into the destination folder
function generatorEngine(type, moduleName, srcPath, destPath, project) {
  // const namespace = stringUtils.classify(`${moduleName}-${type}`);
  const dasherizeName = stringUtils.dasherize(moduleName);
  const classifyName = stringUtils.classify(moduleName);
  const camelizeName = stringUtils.camelize(moduleName);
  const underscoreName = stringUtils.underscore(moduleName);
  const constantName = stringUtils.upper(underscoreName);

  // __DASHERIZE_NAMESPACE__       route-name  =>  route-name
  // __UPPER_DASHERIZE_NAMESPACE__ route-name  =>  ROUTE_NAME
  // __CLASSIFY_NAMESPACE__        route-name  =>  RouteName
  // __CAMEL_NAMESPACE__           route-name  =>  routeName

  return new Promise(function(resolve, reject) {
    gutil.log('installing route');

    gulp
      .src(srcPath)
      .pipe(replace(/__DASHERIZE_NAMESPACE__/g, dasherizeName))
      .pipe(replace(/__CLASSIFY_NAMESPACE__/g, classifyName))
      .pipe(replace(/__UPPER_DASHERIZE_NAMESPACE__/g, constantName))
      .pipe(replace(/__CAMEL_NAMESPACE__/g, camelizeName))
      .pipe(replace(/__PROJECT_NAMESPACE__/g, project ? `@${project}` : ''))
      .pipe(
        rename(function(path) {
          if (path.extname !== '') {
            if (path.basename.includes('.')) {
              // api.service does not have to change its path name
              // everything else will be replaced by its type value
              if (path.basename === 'api.service') {
                path.basename = 'api.service';
              } else {
                const pathArr = path.basename.split('.');
                pathArr[0] = pathArr[0].includes('routing')
                  ? `${dasherizeName}-routing`
                  : dasherizeName;
                path.basename = pathArr.join('.');
              }
            }

            const srcRootPath = project
              ? `projects/${project}/src/app`
              : 'src/app';
            const fullPath =
              path.dirname === '.'
                ? `${srcRootPath}/${dasherizeName}/${path.basename}${
                    path.extname
                  }`
                : `${srcRootPath}/${dasherizeName}/${path.dirname}/${
                    path.basename
                  }${path.extname}`;

            gutil.log(gutil.colors.green('  create'), fullPath);
          }
        })
      )
      .on('error', reject)
      .on('end', routeGeneratorLogger(resolve, { moduleName, project }))
      .pipe(gulp.dest(destPath));
  });
}

// @describe	generate an model,view,store,controller from base template
function runTasks(type, filepath, project) {
  const GENERATOR_PATH = path.join(__dirname, '..', 'skeletons/generators');
  let srcPath = '';
  let destPath = project
    ? path.resolve(`projects/${project}/src/app`)
    : path.resolve('src/app');

  switch (type) {
    case 'route':
      srcPath = [`${GENERATOR_PATH}/${type}/**/*`];
      destPath = path.resolve(destPath, filepath);
      generatorEngine(type, filepath, srcPath, destPath, project);
      break;

    default:
      gutil.log(
        gutil.colors.cyan(`[HINT]`),
        gutil.colors.bold(' valid types: '),
        gutil.colors.cyan(VALID_TYPES.join(', '))
      );
      gutil.log(
        gutil.colors.cyan('[HINT]'),
        'Report this bug to project admin'
      );
  }
}

/**
 * Entry point of the generate command
 * @params type  string, ex: route
 * @params options   string, ex: post
 */
function generate(inputProject, inputType, inputFilepath) {
  let project = inputProject;
  let type = inputType;
  let filepath = inputFilepath;

  if (typeof filepath !== 'string') {
    type = inputProject;
    filepath = inputType;
    project = null;
  }

  // must be inside "angular-cli" project,
  // type must exist and is a string, and path must exist and is a string
  validProjectAndValidUserInput(type, filepath);

  // must be a valid name, must be a valid type, otherwise, exit the program
  validTypesAndValidName(type);

  // only run tasks when it is a valid type and name
  runTasks(type, filepath, project);
}

module.exports = generate;

function validProjectAndValidUserInput(type, filepath) {
  // if the folder 'client/app' is not existed
  // can assume that the project may not be created by Ember Rocks
  if (!fs.existsSync('angular.json')) {
    gutil.log(
      gutil.colors.red(
        `[ERROR] This project may not be created by 'angular-cli'.`
      )
    );
    gutil.log(
      gutil.colors.red(`[ERROR] Unable to find any apps in 'angular-cli.json'`)
    );
    gutil.log(
      gutil.colors.cyan('[HINT] '),
      "Are you at the root path of 'angular-cli' generated project?"
    );
    exitProgram(1);
  }

  // Error out when user did not provide any argument
  if (!type || typeof type !== 'string') {
    gutil.log(
      gutil.colors.red(
        `[ERROR] 'ngg generate' command requires a blueprint name (type) to be specified.`
      )
    );
    gutil.log(
      gutil.colors.cyan(`[HINT]  'ngg generate --help'`),
      'for more detail'
    );
    exitProgram();
  }

  if (!filepath || typeof filepath !== 'string') {
    gutil.log(
      gutil.colors.red(
        `[ERROR] 'ngg generate ${type}' command requires a name to be specified.`
      )
    );
    gutil.log(
      gutil.colors.cyan(`[HINT]  'ngg generate --help'`),
      'for more detail'
    );
    exitProgram();
  }
}

function validTypesAndValidName(type) {
  // Type must be in the `validTypes` array
  if (!VALID_TYPES.includes(type)) {
    gutil.log(
      gutil.colors.red('[ERROR]'),
      gutil.colors.cyan(type),
      gutil.colors.red(' is an invalid blueprint name (type).')
    );

    gutil.log(
      gutil.colors.cyan(`[HINT]`),
      gutil.colors.bold(' valid types:'),
      gutil.colors.cyan(VALID_TYPES.join(', '))
    );

    gutil.log(
      gutil.colors.cyan(`[HINT]  'ngg generate --help'`),
      'for more detail'
    );
    exitProgram();
  }
}

function exitProgram(errNumber) {
  errNumber = errNumber || 0;
  process.exit(errNumber);
}

function routeGeneratorLogger(resolve, options) {
  options = options || {};

  return function() {
    updateRouting(options);

    // gutil.log(
    //   gutil.colors.green('[DONE] A new route'),
    //   gutil.colors.cyan(options.name),
    //   gutil.colors.green('have been successfully created. ('),
    //   gutil.colors.cyan(`src/app/${options.name}`),
    //   gutil.colors.green(')'),
    // );
    resolve();
  };
}
